#!/usr/bin/python

import urllib.request
import json
import os
import getmac
import signal
from threading import Event
import getopt
import sys

from pynovatec.onboarding import onboard
from pynovatec.helpers import write_json_to_file,read_json_from_file
from pynovatec import helpers 
from pynovatec.onboarding.calibration_ops_pumpsense import CalibParamFieldMissingException, CalibException


exit = Event()
product_type=None

fetch_config = {
   'interface': 'br-lan',
   'base_path': '/opt/novatec/data/slave/json_data',
   'sensor_config_file': '/opt/novatec/data/slave/conf/sensor_config.txt',
   'calib_params_file': '/opt/novatec/data/slave/conf/calib_params.txt',
   'frequency':600,
   'api_host': 'api.machinesense.com',
   'product_type':"ps"
}

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr,**kwargs)

def set_api_routes(api_host="api.machinesense.com"):
    #global MACHINE_ID_URL
    #global MACHINE_INFO_URL
    global GAUGE_INFO_URL
    global GAUGE_INFO_URL
    global SENSOR_URL
    global SENSOR_BOARD_URL
    global DATAHUB_URL

    #MACHINE_ID_URL=f'https://{api_host}/api/onboarding/machineid'  # Bad replace
    #MACHINE_INFO_URL=f'https://{api_host}/api/onboarding/machineinfo' # Bad replace

    DATAHUB_URL=f'https://{api_host}/api/datahub'
    GAUGE_INFO_URL=f'https://{api_host}/api/onboarding/gaugeinfo'
    #SENSOR_URL=f'https://{api_host}/api/sensor/11223344001d_VAC_1'
    SENSOR_URL=f'https://{api_host}/api/sensor'
#This is just here for a place holder..  It isn't used and more than likly won't be and will get removed.
    #SENSOR_BOARD_URL='http://40.114.29.93:3000/api/sensorBoard?filter={"where":{"uuid":"b827eba73fbb"},"include":[{"relation":"sensorHW"},{"relation":"sensors","scope":{"include":[{"relation":"sensorType"},{"relation":"sensorAssignments","scope":{"include":{"relation":"machine","scope":{"include":{"relation":"zone"}}}}}]}}]}'
    SENSOR_BOARD_URL=f'http://{api_host}/api/sensorBoard'

#?filter={"where":{"uuid":"b827eba73fbb"},"include":[{"relation":"sensorHW"},{"relation":"sensors","scope":{"include":[{"relation":"sensorType"},{"relation":"sensorAssignments","scope":{"include":{"relation":"machine","scope":{"include":{"relation":"zone"}}}}}]}}]}'

def url_fetch_json(url):
    debug_out("URL: ",url)
    res = urllib.request.urlopen(url)
    if res.getcode() != 200:
        return {}
    contents = res.read()
    body = json.loads(contents.decode("utf-8"))
    debug_out("Body: ",json.dumps(body,indent=4))
    if 'error' in body:
        print("Error fetching: ",url)
        print("Content: ",contents.decode("utf-8"))
        return {}
    return body


def debug_out_enabled(*args, **kwargs):
    print(*args,**kwargs)

def debug_out_disabled(*args, **kwargs):
    pass

debug_out = debug_out_disabled

def fetch_datahub(macAddress):
    global DATAHUB_URL

    #url = f"https://api.machinesense.com/api/datahub?filter[where][macAddress]={macAddress}&filter[include][machines]"
    url = f"{DATAHUB_URL}?filter[where][macAddress]={macAddress}&filter[include][machines]"
    debug_out("url: ",url)
    return url_fetch_json(url)

#def fetch_machine(machine_id):
#    return url_fetch_json(MACHINE_INFO_URL + "/" + machine_id)

def fetch_machine(machineId):
    base_url = f'https://api.machinesense.com/api/machine/{machineId}'
    filter='filter={"include":[{"relation":"models","scope":{"include":{"relation":"assembly","scope":{"include":{"relation":"subAssemblyInstances","scope":{"include":[{"relation":"subAssembly","scope":{"include":{"relation":"collectors"}}},{"relation":"subAssemblyAttributeInstances","scope":{"include":"subAssemblyAttribute"}}]}}}}}},{"relation":"sensorAssignments","scope":{"include":[{"relation":"collectors","scope":{"include":"subAssembly"}},{"relation":"subAssemblyInstance","scope":{"include":[{"relation":"subAssembly"},{"relation":"subAssemblyAttributeInstances","scope":{"include":"subAssemblyAttribute"}}]}}]}},{"relation":"zone","scope":{"include":{"relation":"location","scope":{"include":{"relation":"company"}}}}}]}'
    url=f'{base_url}?{filter}'
    return url_fetch_json(url)

def fetch_gauge(machine_id):
    return url_fetch_json(GAUGE_INFO_URL + "/" + machine_id)

def fetch_qc_param(sensor):
    return url_fetch_json(SENSOR_URL + "/" + sensor)

def fetch_sensor_board(mac):
    uuid = mac.replace(":","")
    filter='{"where":{"uuid":"'+uuid+'"},"include":[{"relation":"sensorHW"},{"relation":"sensors","scope":{"include":[{"relation":"sensorType"},{"relation":"sensorAssignments","scope":{"include":{"relation":"machine","scope":{"include":{"relation":"zone"}}}}}]}}]}'
    return url_fetch_json(SENSOR_BOARD_URL + "?filter=" + filter)

def extract_machine_ids(details):
    machine_ids = []
    for machine in details['machines']:
       machine_id = machine['id']
       machine_ids.append(machine_id)
    return machine_ids

def fetch_gauges(machine_ids):
    gauges=[]
    for machine_id in machine_ids:
       gauge=fetch_gauge(machine_id)
       gauges.append(gauge)
    return gauges 

def fetch_machines(machine_ids):
    machines=[]
    for machine_id in machine_ids:
       machine=fetch_machine(machine_id)
       machines.append(machine)
    return machines

def fetch_qc_params(sensors):
    qc_params=[]
    for sensor in sensors:
        qc_param = fetch_qc_param(sensor)
        qc_params.append(qc_param)
    return qc_params

def extract_sensors(machines):
    sensors=[]
    for machine in machines:
        for sensorAssignment in machine['sensorAssignments']:
            sensors.append(sensorAssignment['sensorId'])
    return sensors

def filter_list(items,filter):
    filtered_list=[]
    for item in items:
       if filter in item:
           filtered_list.append(item)
    return filtered_list    

#contents = urllib.request.urlopen(MACHINE_ID_URL + "/" + MAC).read()
#body = json.loads(contents.decode("utf-8"))
#print(json.dumps(body,indent=4))

def fetch_onboard():

    #dh_details=url_fetch_json(MACHINE_ID_URL + "/" + MAC)
    #print(json.dumps(dh_details,indent=4))
    dh_details = fetch_datahub(MAC)
    if type(dh_details) == list:
        if (len(dh_details) > 0):
       	    dh_details = dh_details[0]
        else: 
            return {}

    if not dh_details:
        return {}
    if 'lastSeen' in dh_details:
        del dh_details['lastSeen']
    if 'modifiedOn' in dh_details:
        del dh_details['modifiedOn']

    # Kludge for testing to getaround mismatched mac address..

    #dh_details['macAddress']="dc:a6:32:ec:71:fe"

    #print(dh_details)
    machine_ids = extract_machine_ids(dh_details)
    gauges = fetch_gauges(machine_ids)
    #print("Gauges: ",gauges)
    machines = fetch_machines(machine_ids)
    sensors=extract_sensors(machines)
    vac_sensor_list=filter_list(sensors,"VAC_1")
    ps_sensor_list=filter_list(sensors,"PS_1")
    sensor_list = vac_sensor_list + ps_sensor_list
    qc_params = fetch_qc_params(sensor_list)
    onboard_json = {
        'type' : "ps",
        'dh_details': dh_details,
        'gauges': gauges,
        'machines': machines,
        'qc_params': qc_params

    }
    return onboard_json

def perform_onboard(onboard_json):
    if not onboard_json:
        return
    try:
        onboard(onboard_json,**fetch_config)
    except CalibException as e: # This is stull successful.                                  
        missing_calibs=True                                                                  
        error_message=str(e)                                                                 
    except Exception as e:
        print("perform_onboard - Uncaught exception: ",str(e))
        write_json_to_file("/tmp/raw_onboard.json",onboard_json)
        raise e
        return

    write_json_to_file(raw_file,onboard_json)



def extract_pa_machine_ids(dh_details):
    machine_ids=[]

    sensor_assigns = dh_details.get("sensorAssignments",None)
    if sensor_assigns:
        machine_ids.append(sensor_assigns.get("machineId",None))
    return machine_ids


def fetch_pa_dh_details():
    stripped_mac = MAC.replace(":","").lower()
    debug_out("stripped mac: ",stripped_mac)
    url=f"https://api.machinesense.com/api/sensor/{stripped_mac}_PF_1?filter[include][sensorAssignments]=machine"    
    dh_details=url_fetch_json(url)
    debug_out("DH Details: ",dh_details)
    if type(dh_details) == list:
        dh_details = dh_details[0]

    if 'lastSeen' in dh_details:
        del dh_details['lastSeen']
    if 'modifiedOn' in dh_details:
        del dh_details['modifiedOn']

    #Kludge to align with pumpsense.
    dh_details['machines']=[]
    sa = dh_details.get('sensorAssignments',None)
    if sa is None:
        debug_out("No Sensor Assignements")
        raise Exception("No Sensor Assignments.")
    machine = dh_details.get('sensorAssignments').get('machine',None)
    if machine is None:
        debug_out("No Machine information..")
    dh_details['machines'].append(machine)
    if "macAddress" not in dh_details:
        dh_details['macAddress'] = MAC
    #End kludge
    return dh_details

def fetch_plc_dh_details():
    stripped_mac = MAC.replace(":","").lower()
    debug_out("stripped mac: ",stripped_mac)
    url=f"https://api.machinesense.com/api/sensor/{stripped_mac}_PLC_1?filter[include][sensorAssignments]=machine"    
    dh_details=url_fetch_json(url)
    debug_out("DH Details: ",dh_details)
    if type(dh_details) == list:
        dh_details = dh_details[0]

    if not dh_details:
        debug_out("No dh_details")
        return dh_details
    if 'lastSeen' in dh_details:
        del dh_details['lastSeen']
    if 'modifiedOn' in dh_details:
        del dh_details['modifiedOn']

    #Kludge to align with pumpsense.
    dh_details['machines']=[]
    machine = dh_details.get('sensorAssignments').get('machine')
    dh_details['machines'].append(machine)
    if "macAddress" not in dh_details:
        dh_details['macAddress'] = MAC
    #End kludge
    return dh_details

def fetch_pa_onboard():
    debug_out("Fetching pa onboard for mac: ",MAC)
    #stripped_mac = MAC.replace(":","").lower()
    #print("stripped mac: ",stripped_mac)
    #url=f"https://api.machinesense.com/api/sensor/{stripped_mac}_PF_1?filter[include][sensorAssignments]=machine"    
    #dh_details=url_fetch_json(url)
    #print("DH Details: ",dh_details)
    #if type(dh_details) == list:
    #    dh_details = dh_details[0]
    dh_details = fetch_pa_dh_details()
    if not dh_details:
        return {}
    machine_ids = extract_pa_machine_ids(dh_details)
    debug_out("machine_ids: ",machine_ids)
    gauges = fetch_gauges(machine_ids)
    machines = fetch_machines(machine_ids)
    debug_out("machines: ",json.dumps(machines,indent=4))
    sensors=extract_sensors(machines)
    debug_out("Sensors: ",sensors)
    #vac_sensor_list=filter_list(sensors,"VAC_1")
    qc_params = fetch_qc_params(sensors)
    #'qc_params': qc_params
    debug_out("********** FETCH SENSOR BOARD ******************")
    sensorBoard = fetch_sensor_board(MAC)
    debug_out("************** END FETCH ***********************")
    onboard_json = {
        'type': "pa",
        'dh_details': dh_details,
        'gauges': gauges,
        'machines': machines,
	'qc_params': qc_params,
	'sensor_boards': sensorBoard

    }
    return onboard_json

def fetch_plc_onboard():
    debug_out("Fetching pa onboard for mac: ",MAC)
    dh_details = fetch_plc_dh_details()
    if not dh_details:
        return {}
    machine_ids = extract_pa_machine_ids(dh_details)
    debug_out("machine_ids: ",machine_ids)
    gauges = fetch_gauges(machine_ids)
    machines = fetch_machines(machine_ids)
    debug_out("machines: ",json.dumps(machines,indent=4))
    sensors=extract_sensors(machines)
    debug_out("Sensors: ",sensors)
    #vac_sensor_list=filter_list(sensors,"VAC_1")
    qc_params = fetch_qc_params(sensors)
    #'qc_params': qc_params
    debug_out("********** FETCH SENSOR BOARD ******************")
    sensorBoard = fetch_sensor_board(MAC)
    debug_out("************** END FETCH ***********************")
    onboard_json = {
        'type': "plc",
        'dh_details': dh_details,
        'gauges': gauges,
        'machines': machines,
	'qc_params': qc_params,
	'sensor_boards': sensorBoard

    }
    return onboard_json


def do_onboarding(fetch_onboard_fun):
    try:
        onboard_json = fetch_onboard_fun()
        if not helpers.comparers.file_compare(raw_file,onboard_json):
            debug_out("New json file, doing onboard")
            perform_onboard(onboard_json)
    except Exception as e:
        eprint("Exception doing onboard: ",str(e))
        

def try_ps_onboard():
    do_onboarding(fetch_onboard)
    #onboard_json = fetch_onboard()
    #if not helpers.comparers.file_compare(raw_file,onboard_json):
    #    debug_out("New json file, doing onboard")
    #    perform_onboard(onboard_json)

def try_pa_onboard():
    do_onboarding(fetch_pa_onboard)
    #onboard_json = fetch_pa_onboard()
    #if not helpers.comparers.file_compare(raw_file,onboard_json):
    #    debug_out("New json file, doing onboard")
    #    perform_onboard(onboard_json)

def try_plc_onboard():
    do_onboarding(fetch_plc_onboard)
    #onboard_json = fetch_plc_onboard()
    #if not helpers.comparers.file_compare(raw_file,onboard_json):
    #    debug_out("New json file, doing onboard")
    #    perform_onboard(onboard_json)
    #    #print("PLC ONBOARD: ",json.dumps(onboard_json,indent=4))



keep_running = True

def signal_to_quit(signum, frame):
    global keep_running

    exit.set()
    keep_running = False

def signal_to_do_nothing(signum, frame):
    exit.set()
    pass

def main_loop(frequency=600):
    while keep_running :
        try: 
            try_onboard()
        except Exception as e:
             print("Exception doing onboard: ",str(e))
             raise e
        exit.wait(frequency)
        exit.clear()

config_file = None


def usage(prog):
    print(f"Usage: {prog} [-c file] [-m mac]")
    sys.exit()

#def set_onboard_config():
#    global onboard_config
#
#    onboard_config = {
#       'interface': fetch_config['interface'],
#       'base_path': fetch_config['base_path'],
#       'sensor_config_file': fetch_config['sensor_config_file'],
#       'calib_params_file': fetch_config['calib_params_file']
#    }



def main():
#    global onboard_config
    global fetch_config
    global MAC
    global raw_file
    global config_file
    global product_type
    global try_onboard
    global debug_out
    tmp_mac = None

    try:
        opts, args = getopt.getopt(sys.argv[1:],"c:m:t:d")
    except getopt.GetoptError:
        print("Error: Invalid argument ")
        usage(__file__)
    for opt,arg in opts:
        #print("opt is: ",opt)
        if opt == "-c":
            config_file = arg
        if opt == "-m":
            tmp_mac = arg.lower().strip()
        if opt == "-t":
            product_type = arg
        if opt == "-d":
            print("DEBUG ENABLE")
            debug_out = debug_out_enabled

    if config_file is not None:
        tmp_config = read_json_from_file(config_file)
        fetch_config = {**fetch_config,**tmp_config}
    if product_type is None:
        product_type = fetch_config['product_type']
    MAC = getmac.get_mac_address(interface=fetch_config['interface']).strip()
    if tmp_mac is not None:
        MAC = tmp_mac
        fetch_config['mac'] = MAC
    set_api_routes(fetch_config['api_host'])

    if product_type == 'pa':
        try_onboard = try_pa_onboard
    elif product_type == 'plc':
        try_onboard = try_plc_onboard
    else:
        try_onboard = try_ps_onboard
    #set_onboard_config()
    raw_path = os.path.join(fetch_config['base_path'],"raw")
    raw_file = os.path.join(raw_path,"onboard.json")
    main_loop(fetch_config['frequency'])


if __name__ == '__main__':
    signal.signal(signal.SIGINT,signal_to_quit)
    signal.signal(signal.SIGTERM,signal_to_quit)
    signal.signal(signal.SIGQUIT,signal_to_quit)
    signal.signal(signal.SIGHUP,signal_to_do_nothing)	# Make it so SIGHUP breaks the wait.
    main()


